use std::sync::Arc;

use axum::{
    extract::Path,
    http::StatusCode,
    response::IntoResponse,
    routing::{get, post},
    Json, Router,
};
use serde::{Deserialize, Serialize};
use sqlx::PgPool;

use crate::{
    config::Config,
    db::{AvailableDb, Database},
    jwt::make_jwt,
    utils::internal_error,
    worker::{Message, Queue},
    AppCtx,
};

#[derive(Debug, Clone, Serialize)]
struct JwtResponse {
    token: String,
}

#[derive(Debug, Clone, Deserialize)]
struct LoginRequest {
    username: String,
}

pub fn routes() -> Router<AppCtx> {
    Router::new()
        .route("/finish/:token", get(handle_login))
        .route("/generate", post(start_login))
}

async fn start_login(
    q: Queue,
    Database(mut conn): Database,
    Json(LoginRequest { username }): Json<LoginRequest>,
) -> Result<(), (StatusCode, String)> {
    let user = conn
        .get_or_create_user_by_name(&username)
        .await
        .map_err(internal_error)?;

    let token = conn
        .create_magic_token(&user.id)
        .await
        .map_err(internal_error)?;

    q.send(Message::SendToken {
        user: user.username,
        token: token,
    })
    .await;

    Ok(())
}

async fn handle_login(
    Path(token): Path<String>,
    cfg: Arc<Config>,
    Database(mut conn): Database,
) -> Result<Json<JwtResponse>, (StatusCode, String)> {
    let user = conn.check_magic_token(&token).await.map_err(|e| {
        tracing::error!("Error logging in with token '{token}': ");
        (StatusCode::UNAUTHORIZED, "Invalid Token".to_owned())
    })?;

    let jwt = make_jwt(&*cfg, &user.to_string()).map_err(internal_error)?;

    Ok(Json(JwtResponse { token: jwt }))
}
