mod model;
pub mod postgres;
pub mod surreal;

use crate::error::Result;
use async_trait::async_trait;
use axum::{
    extract::FromRequestParts,
    http::{request::Parts, StatusCode},
};
use sqlx::{types::Uuid, PgPool};
use surrealdb::{engine::any::Any, Surreal};
use tracing::error;

use crate::AppCtx;

use self::model::{DbUser, NewUser};

#[derive(Clone, Debug)]
pub enum AvailableDb {
    Postgres(PgPool),
}

#[async_trait]
pub trait DatabaseAdapter {
    async fn get_user_by_name(&mut self, name: &str) -> Result<DbUser>;
    async fn get_or_create_user_by_name(&mut self, name: &str) -> Result<DbUser>;
    async fn create_user(&mut self, name: NewUser) -> Result<DbUser>;
    async fn create_track(&self);
    async fn create_magic_token(&mut self, user_id: &Uuid) -> Result<String>;
    async fn check_magic_token(&mut self, token: &str) -> Result<Uuid>;
}
pub struct Database(pub Box<dyn DatabaseAdapter + Send + Sync + 'static>);

#[async_trait]
impl FromRequestParts<AppCtx> for Database {
    type Rejection = (StatusCode, String);

    async fn from_request_parts(
        _parts: &mut Parts,
        state: &AppCtx,
    ) -> Result<Self, Self::Rejection> {
        match &state.db {
            AvailableDb::Postgres(pool) => {
                let conn = pool.acquire().await.map_err(|e| {
                    error!("{e}");
                    (
                        StatusCode::INTERNAL_SERVER_ERROR,
                        "Internal Server Error".to_owned(),
                    )
                })?;

                Ok(Self(Box::new(conn)))
            }
        }
    }
}

// #[async_trait]
// impl FromRequestParts<Surreal<Any>> for DatabaseConnection {
//     type Rejection = (StatusCode, String);

//     async fn from_request_parts(
//         _parts: &mut Parts,
//         state: &Surreal<Any>,
//     ) -> Result<Self, Self::Rejection> {
//         let conn = Surreal::from_ref(state);

//         // let conn = pool.acquire().await.map_err(|e| {
//         //     error!("{e}");
//         //     (
//         //         StatusCode::INTERNAL_SERVER_ERROR,
//         //         "Internal Server Error".to_owned(),
//         //     )
//         // })?;

//         Ok(Self(Box::new(conn)))
//     }
// }
