use lazy_static::lazy_static;
use rand::{distributions::Alphanumeric, prelude::*};
use std::{ops::DerefMut, str::FromStr};

use crate::error::Result;

use async_trait::async_trait;
use sqlx::{pool::PoolConnection, query, query_as, types::Uuid, PgPool, Postgres};

use super::{
    model::{DbUser, NewUser},
    DatabaseAdapter,
};

pub async fn connect(url: impl AsRef<str>) -> Result<PgPool> {
    let pool = sqlx::PgPool::connect(url.as_ref()).await?;
    sqlx::migrate!().run(&mut pool.acquire().await?).await?;
    Ok(pool)
}

#[async_trait]
impl DatabaseAdapter for PoolConnection<Postgres> {
    async fn get_user_by_name(&mut self, name: &str) -> Result<DbUser> {
        let res = query_as!(
            DbUser,
            "SELECT * FROM Person WHERE username = $1 LIMIT 1",
            name
        )
        .fetch_one(self.deref_mut())
        .await?;

        Ok(res)
    }
    async fn get_or_create_user_by_name(&mut self, name: &str) -> Result<DbUser> {
        match query_as!(
            DbUser,
            "SELECT * FROM Person WHERE username = $1 LIMIT 1",
            name
        )
        .fetch_one(self.deref_mut())
        .await
        {
            Err(sqlx::Error::RowNotFound) => {
                let Some((displayname, host)) = name.split_once('@') else {
                    return Err(sqlx::Error::RowNotFound.into());
                };
                self.create_user(NewUser {
                    username: name.to_string(),
                    host: host.to_string(),
                    name: Some(displayname.to_string()),
                    is_admin: false,
                })
                .await
            }
            other => Ok(other?),
        }
    }
    async fn create_user(&mut self, user: NewUser) -> Result<DbUser> {
        Ok(query_as!(
            DbUser,
            "INSERT INTO Person VALUES (gen_random_uuid(), $1, $2, $3, $4) RETURNING *",
            user.username,
            user.name,
            user.is_admin,
            user.host,
        )
        .fetch_one(self.deref_mut())
        .await?)
    }

    async fn create_track(&self) {}
    async fn create_magic_token(&mut self, user_id: &Uuid) -> Result<String> {
        let token: String = thread_rng()
            .sample_iter(Alphanumeric)
            .take(64)
            .map(char::from)
            .collect();

        query!(
            "INSERT INTO Magic VALUES ($1, $2, now(), '15 min')",
            token,
            user_id
        )
        .execute(self.deref_mut())
        .await?;
        Ok(token)
    }
    async fn check_magic_token(&mut self, token: &str) -> Result<Uuid> {
        let user = query!(
            "SELECT id FROM Magic WHERE token = $1 AND now() - created_at < max_age ",
            token
        )
        .fetch_one(self.deref_mut())
        .await?;

        Ok(user.id)
    }
}
