use std::{
    alloc::System,
    time::{Duration, SystemTime},
};

use anyhow::anyhow;
use josekit::{
    jws::{JwsHeader, HS256},
    jwt::{self, JwtPayload, JwtPayloadValidator},
    JoseError,
};
use sqlx::types::Uuid;

use crate::config::Config;

pub fn make_jwt(cfg: &Config, user: &str) -> Result<String, JoseError> {
    let signer = HS256.signer_from_bytes(cfg.hmac_key.as_bytes())?;

    let mut header = JwsHeader::new();
    header.set_token_type("JWT");

    let mut payload = JwtPayload::new();
    payload.set_subject(user);
    payload.set_expires_at(&(SystemTime::now() + Duration::from_secs(432000))); // expires in 5 days
    payload.set_issued_at(&SystemTime::now());
    payload.set_audience(vec!["user_token"]);

    let token = jwt::encode_with_signer(&payload, &header, &signer)?;

    Ok(token)
}

pub fn verify_jwt(cfg: &Config, token: &str) -> Result<Uuid, anyhow::Error> {
    let verifier = HS256.verifier_from_bytes(cfg.hmac_key.as_bytes())?;

    let (payload, header) = jwt::decode_with_verifier(token, &verifier)?;
    let mut validator = JwtPayloadValidator::new();
    validator.set_base_time(SystemTime::now());
    validator.validate(&payload)?;

    let Some(id) = payload.subject() else {
        return Err(anyhow!("Missing Subject claim"));
    };

    Ok(Uuid::try_parse(id)?)
}
