use std::sync::Arc;

use axum::{extract::Query, http::StatusCode, routing::get, Json, Router};
use config::{make_config, Config};
use db::{AvailableDb, Database};
use error::Error;
use serde::{Deserialize, Serialize};
use tower_http::services::{ServeDir, ServeFile};
use tracing::{error, info};
use tracing_subscriber::prelude::*;
use utils::internal_error;
use worker::Queue;

mod api;
mod config;
mod db;
mod error;
mod jwt;
mod utils;
mod worker;

#[derive(Clone, Debug)]
pub struct AppCtx {
    db: AvailableDb,
    q: Queue,
    config: Arc<Config>,
}

#[tokio::main]
async fn main() {
    // init logging
    tracing_subscriber::registry()
        .with(tracing_subscriber::EnvFilter::from_default_env())
        .with(tracing_subscriber::fmt::layer())
        .init();

    let q = worker::start().await.expect("Worker");

    // let db_conn = db::surreal::connect("localhost:8000").await.expect("DB");
    let db_conn = db::postgres::connect("postgres://postgres:password@localhost")
        .await
        .expect("db connection");

    let router = Router::new()
        .nest("/api", api::routes())
        .route("/.well-known/webfinger", get(webfinger))
        .with_state(AppCtx {
            db: db::AvailableDb::Postgres(db_conn),
            config: Arc::new(make_config().expect("config")),
            q,
        })
        .fallback_service(
            ServeDir::new("public").not_found_service(ServeFile::new("public/index.html")),
        );

    match tokio::net::TcpListener::bind("0.0.0.0:8675").await {
        Ok(socket) => {
            info!("Starting server on {}", socket.local_addr().expect("addr"));
            axum::serve(socket, router)
                .with_graceful_shutdown(shutdown())
                .await
                .expect("Error running server");
        }
        Err(e) => {
            error!("Error binding address: {e}");
            return;
        }
    }
}

async fn shutdown() {
    tokio::signal::ctrl_c().await.expect("sigint listener");
}

#[derive(Debug, Clone, Deserialize)]
struct WFQuery {
    resource: String,
}
/// WebFinger result that may serialized or deserialized to JSON
#[derive(Debug, Serialize, Deserialize, PartialEq)]
pub struct Webfinger {
    /// The subject of this WebFinger result.
    ///
    /// It is an `acct:` URI
    pub subject: String,

    /// A list of aliases for this WebFinger result.
    #[serde(default)]
    pub aliases: Vec<String>,

    /// Links to places where you may find more information about this resource.
    pub links: Vec<Link>,
}

/// Structure to represent a WebFinger link
#[derive(Debug, Serialize, Deserialize, PartialEq)]
pub struct Link {
    /// Tells what this link represents
    pub rel: String,

    /// The actual URL of the link
    #[serde(skip_serializing_if = "Option::is_none")]
    pub href: Option<String>,

    /// The Link may also contain an URL template, instead of an actual URL
    #[serde(skip_serializing_if = "Option::is_none")]
    pub template: Option<String>,

    /// The mime-type of this link.
    ///
    /// If you fetch this URL, you may want to use this value for the Accept header of your HTTP
    /// request.
    #[serde(rename = "type", skip_serializing_if = "Option::is_none")]
    pub mime_type: Option<String>,
}

async fn webfinger(
    Query(q): Query<WFQuery>,
    Database(mut conn): Database,
) -> Result<Json<Webfinger>, (StatusCode, String)> {
    let (_, name) = q
        .resource
        .split_once(':')
        .ok_or_else(|| (StatusCode::BAD_REQUEST, "Malformed Resource".to_owned()))?;

    match conn.get_user_by_name(name).await {
        Err(Error::DatabaseError(sqlx::Error::ColumnNotFound(_))) => {
            Err((StatusCode::NOT_FOUND, "Unknown user".into()))
        }
        Err(e) => Err(internal_error(e)),
        Ok(user) => Ok(Json(Webfinger {
            subject: q.resource,
            aliases: vec![user.username.clone(), user.displayname.unwrap_or_default()],
            links: vec![Link {
                rel: "http://webfinger.net/rel/profile-page".to_string(),
                href: format!("http://localhost:8675/{}", user.username).into(),
                template: None,
                mime_type: None,
            }],
        })),
    }
}
