use std::{
    collections::VecDeque,
    sync::Arc,
    time::{Duration, SystemTime},
};

use async_trait::async_trait;
use axum::{
    extract::FromRequestParts,
    http::{request::Parts, StatusCode},
};
use celery::Celery;
use tokio::{
    spawn,
    sync::{Mutex, Notify},
    task::spawn_blocking,
};
use tracing::{debug, error, info, warn};

use crate::AppCtx;

pub mod ap;

#[derive(Debug, Clone)]
pub struct Queue {
    q: Arc<Mutex<VecDeque<Task>>>,
    notify: Arc<tokio::sync::Notify>,
}

impl Queue {
    pub async fn send(&self, msg: Message) {
        self.q.lock().await.push_back(Task::new(msg));
        self.notify.notify_one();
    }
}

#[async_trait]
impl FromRequestParts<AppCtx> for Queue {
    type Rejection = StatusCode;

    async fn from_request_parts(
        _parts: &mut Parts,
        state: &AppCtx,
    ) -> Result<Self, Self::Rejection> {
        Ok(state.q.clone())
    }
}

#[derive(Debug, Clone)]
pub enum Message {
    SendToken { user: String, token: String },
}

#[derive(Debug, Clone)]
struct Task {
    tries: u16,
    delay: Duration,
    last_attempt: Option<SystemTime>,
    msg: Message,
}

impl Task {
    fn new(msg: Message) -> Self {
        Self {
            tries: 0,
            delay: Duration::ZERO,
            last_attempt: None,
            msg,
        }
    }
}

pub async fn start() -> Result<Queue, anyhow::Error> {
    info!("Starting worker task");
    // let app = celery::app!(
    //     broker = AMQPBroker { "amqp://127.0.0.1:5672".to_string() },
    //     tasks = [
    //         send_token
    //     ],
    //     task_routes = [
    //         "send_token" => "outbound",
    //         "*" => "celery",
    //     ],
    //     prefetch_count = 2,
    //     heartbeat = Some(10),
    // )
    // .await?;

    let q = Arc::new(Mutex::new(VecDeque::<Task>::new()));
    let notify = Arc::new(Notify::new());

    let consumer = q.clone();
    let listener = notify.clone();

    spawn(async move {
        loop {
            listener.notified().await;

            if let Some(task) = consumer.lock().await.pop_front() {
                debug!("Recieved message {:?}", task.msg);
                if let Err(e) = dispatch(&task.msg).await {
                    error!("Worker errorL {e}");
                    consumer.lock().await.push_back(task);
                }
            } else {
                warn!("Notified on empty queue");
            }
        }
    });

    // let consumer = app.clone();
    // spawn(async move {
    //     if let Err(e) = consumer.consume_from(&["celery", "outbound"]).await {
    //         error!("Worker error: {e}");
    //     }
    // });

    Ok(Queue { q, notify })
}

async fn dispatch(msg: &Message) -> anyhow::Result<()> {
    match msg {
        Message::SendToken { user, token } => ap::send_token(user, token).await?,
    }
    Ok(())
}
